# TrRMIte

TrRMIte implements **event-based RPC** _(Remote Procedure Calls)_ on top of Ruby's [`DRb` facilities](https://ruby-doc.org/stdlib/libdoc/drb/rdoc/index.html); it's main characteristics are:

- transactional, non-blocking event ingestion
- very low-latency event broking
- "exactly once" event delivery

Unlike other event brokers, it is very opinionated:

- it doesn't offer a choice between **transactional** and **batch** event ingestion
- nor does it offer a choice between **blocking** and **non-blocking** ingestion
- it doesn't offer a choice between **optimising for latency** and **optimising for throughput**
- it doesn't offer a choice between various event delivery semantics (**"at least once"**, **"at most once",** **"exactly once"**)

Instead it makes opinionated choices in each of these areas based on the requirements for building a real-time telephony platform.

As a consequence, it does not implement the following features offered by more mainstream event brokers, things like:

- event log retention
- event log compaction
- SQL-like event log querying

The tooling around the event log in each of these areas is more basic: `cron`-based daily log rotation and backup, `gzip`-only log compaction, command-line log querying with things like `jq`, `grep`, `sed` and `awk`.

### What's in a name?

TrRMIte stands for "**Tr**ansactional **RMI** _(Remote Method Invocation)_ for **te**lephony"

### I thought you said RPC, not RMI?

In the [object-oriented programming](https://en.wikipedia.org/wiki/Object-oriented_programming) paradigm, **RPC** calls are represented by [remote method invocation](https://en.wikipedia.org/wiki/Remote_method_invocation) (**RMI**).

### Hmmm… I still think you're just making it up!

Notable RPC implementations and analogues include:

- [Distributed Ruby](https://en.wikipedia.org/wiki/Distributed_Ruby) (DRb) allows Ruby programs to communicate with each other on the same machine or over a network. DRb uses remote method invocation (RMI) to pass commands and data between processes.
- [Elixir](https://en.wikipedia.org/wiki/Elixir_%28programming_language%29) builds on top of the Erlang VM and allows process communication _(Elixir/Erlang processes, not OS processes)_ within the same cluster out-of-the-box via built-in mailboxes. Each Elixir process has its own mailbox. A process communicates with another by sending a message to the target's mailbox.

The ["Remote procedure call" Wikipedia page](https://en.wikipedia.org/wiki/Remote_procedure_call) clarifies a lot the above in more detail.

## Installation

Add this line to your application's Gemfile:

```ruby
gem 'TrRMIte'
```

And then execute:

    $ bundle

Or install it yourself as:

    $ gem install TrRMIte

## Getting started

TrRMIte and its various utilities require a couple of directories to exist in your environment.  All of them are fully configurable, but the quickest way to get started is to pre-create the default ones _(and then forget about them)_:

    mkdir -p $(brew --prefix)/var/lib/trrmite
    mkdir -p $(brew --prefix)/var/lib/trrmite/trrmite.backups

## Usage

### quiet mode

In development, it's easiest to run TrRMIte in the foreground, in so-called quiet mode:

```
$ exe/trrmite
```

You can use `CTRL-C` to interrupt it right there and then.

Or you can run it in the background:

```
$ exe/trrmite &
```

In this case you can kill it using your shell's job control commands.

### verbose mode

If you prefer to see the status line, you can run TrRMIte in verbose mode:

```
$ exe/trrmite -verbose
```

This will display the TrRMIte status line in the console, and it will update the status line in real-time as the state of the broker changes.

### broker state

if you want to check the current state of the broker, even when running in quiet mode, then you use the TrRMIte status script:

```
$ exe/trrmite_status
```

This will print a one-off status line to `STDOUT`, typically the console you are running TrRMIte in.  If you want to pretty print it, you can run:

```
$ exe/trrmite_status | jq .
```

Pretty-printed JSON FTW!

If you want to check the current state of the broker against a bundled version of TrRMIte, then you can use:

```
$ bundle exec trrmite_status
```

... and pipe it into `jq` for pretty-printing!

### the `trrmite` status line

Because the TrRMIte status line is printed as JSON, you can pipe the `trrmite` output into `jq` and have all broker state changes pretty-printed to the console:

```
$ exe/trrmite -verbose | jq .
```

_(or else pipe it into `jq -c .` for compact, one-line pretty-printed JSON)_

If you prefer to have the status line updated each time the broker status changes _(instead of continuously appending them)_, you can run the following pipeline:

```
$ ./exe/trrmite -verbose | while read status_line ; do clear ; echo "${status_line}" | jq . ; done
```

This is probably the best way to run `trrmite` in development, even though the above pipeline will run a new instance of `jq` for each and every line of TrRMIte output.  If that is a concern/problem, the following approach will combine the best of both worlds:

```
[ -p /tmp/jqpipe ] || mkfifo /tmp/jqpipe ; \
( cat > /tmp/jqpipe ) & ; \
( jq . < /tmp/jqpipe) & ; \
exe/trrmite -verbose | \
while read status_line ; do clear ; echo "${status_line}" > /tmp/jqpipe; done
```

Named pipes FTW!

### using the `trrmite_control` script

In deployed environments, it's easiest to run TrRMIte as a daemon using the corresponding control script:

    $ exe/trrmite_control

This control script accepts the typical commands provided by the `daemons` gem:

```
Usage: trrmite_control <command> <options> -- <application options>

* where <command> is one of:
  start         start an instance of the application
  stop          stop all instances of the application
  restart       stop all instances and restart them afterwards
  reload        send a SIGHUP to all instances of the application
  run           start the application and stay on top
  zap           set the application to a stopped state
  status        show status (PID) of application instances

* and where <options> may contain several of the following:

    -t, --ontop                      Stay on top (does not daemonize)
    -s, --shush                      Silent mode (no output to the terminal)
    -f, --force                      Force operation
    -n, --no_wait                    Do not wait for processes to stop

Common options:
    -h, --help                       Show this message
        --version                    Show version
```

The output, log and PID file for the daemon will be written to the `/var/lib/trrmite/` directory (on Linux) or the  `/usr/local/var/lib/trrmite/` directory (on Mac).

## Producers & Consumers

### Publish Events

Various ways to **publish** events are provided in [`publish_events.rb`](./contrib/publish_events.rb)

### Schedule Events

Various ways to **schedule** events are provided in [`schedule_events.rb`](./contrib/schedule_events.rb)

### Consume Events

Various ways to **consume** events are provided in [`consume_events.rb`](./contrib/consume_events.rb)

## High-Level Overview

![TrRMIte High-Level Overview](TrRMIte.png)

## Development

After checking out the repo, run `bin/setup` to install dependencies. Then, run `rake test` to run the tests. You can also run `bin/console` for an interactive prompt that will allow you to experiment.

To install this gem onto your local machine, run `bundle exec rake install`. To release a new version, update the version number in `version.rb`, and then run `bundle exec rake release`, which will create a git tag for the version, push git commits and tags, and push the `.gem` file to [rubygems.org](https://rubygems.org).

## Debugging

The TrRMIte broker and all TrRMIte consumers are highly multi-threaded applications;  during development, exceptions raised in any of these threads may cause the thread to silently terminate, causing hard-to-debug issues.

By setting the `abort_on_exception` flag, exceptions raised in any running thread will cause Ruby to abort execution and print the exception _(incl. stacktrace)_ to the console, making it much easier to pinpoint the root cause.

Just stick the following snippet somewhere appropriate in the broker and/or consumer scripts:

`Thread.abort_on_exception = true`

The [Ruby documentation for the `Thread` class](http://ruby-doc.org/core-2.0.0/Thread.html#method-c-abort_on_exception-3D) has more detail.

## Contributing

Bug reports and pull requests are welcome on GitHub at https://github.com/simplybusiness/TrRMIte. This project is intended to be a safe, welcoming space for collaboration, and contributors are expected to adhere to the [Contributor Covenant](http://contributor-covenant.org) code of conduct.


## License

The gem is available as open source under the terms of the [MIT License](http://opensource.org/licenses/MIT).

